from __future__ import annotations
from abc import ABC, abstractmethod
import time
from typing import ClassVar, Dict, Optional, Set, Type
from structlog import BoundLogger

from zenx.settings import Settings



class DBClient(ABC):
    name: ClassVar[str]
    _registry: ClassVar[Dict[str, Type[DBClient]]] = {}
    

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "name"):
            raise TypeError(f"DBClient subclass {cls.__name__} must have a 'name' attribute.")
        cls._registry[cls.name] = cls

    
    @classmethod
    def get_db(cls, name: str) -> Type[DBClient]:
        if name not in cls._registry:
            raise ValueError(f"DBClient '{name}' is not registered. Available db clients: {list(cls._registry.keys())}")
        return cls._registry[name]

    
    def __init__(self, logger: BoundLogger, settings: Settings) -> None:
        self.logger = logger
        self.settings = settings

        
    @abstractmethod
    async def start(self) -> None:
        ...

    
    @abstractmethod
    async def _connect(self) -> None:
        ...


    @abstractmethod
    async def insert(self, id: str, spider_name: str) -> None:
        ...


    @abstractmethod
    async def exists(self, id: str, spider_name: str) -> bool:
        ...


    @abstractmethod
    async def delete(self, id: str, spider_name: str) -> None:
        ...


    @abstractmethod
    async def cleanup(self, days: int) -> None:
        ...


    @abstractmethod
    async def close(self) -> None:
        ...


class MemoryDB(DBClient):
    name = "memory"
    processed_ids_set: Set = set()
    

    async def start(self) -> None:
        pass


    async def _connect(self) -> None:
        pass


    async def insert(self, id: str, spider_name: str) -> None:
        unique_id = f"{spider_name}_{id}"
        self.processed_ids_set.add(unique_id)
        self.logger.debug("inserted", id=unique_id, db=self.name)


    async def exists(self, id: str, spider_name: str) -> bool:
        unique_id = f"{spider_name}_{id}"
        return unique_id in self.processed_ids_set


    async def delete(self, id: str, spider_name: str) -> None:
        unique_id = f"{spider_name}_{id}"
        self.processed_ids_set.remove(unique_id)
        self.logger.debug("deleted", id=unique_id, db=self.name)


    async def cleanup(self, days: int) -> None:
        pass


    async def close(self) -> None:
        pass


try:
    import redis.asyncio as redis

    class RedisDB(DBClient): # type: ignore [reportRedeclaration]
        name = "redis"
        processed_ids_zset: str = "processed:ids:zset"
        
        
        def __init__(self, logger: BoundLogger, settings: Settings) -> None:
            super().__init__(logger, settings)
            self.r: Optional[redis.Redis] = None

    
        async def start(self) -> None:
            try:
                await self._connect()
            except Exception:
                self.logger.exception("db_start_failed", db=self.name)
                raise


        async def _connect(self) -> None:
            self.logger.debug("connecting", db=self.name)
            self.r = redis.Redis(
                host=self.settings.DB_HOST,
                port=self.settings.DB_PORT,
                password=self.settings.DB_PASS,
                socket_timeout=5,
                socket_connect_timeout=5,
                decode_responses=True,
            )
            await self.r.ping()
            self.logger.debug("connected", db=self.name)


        async def insert(self, id: str, spider_name: str) -> None:
            unique_id = f"{spider_name}_{id}"
            timestamp = int(time.time())
            await self.r.zadd(self.processed_ids_zset, {unique_id: timestamp})
            self.logger.debug("inserted", id=unique_id, db=self.name)


        async def exists(self, id: str, spider_name: str) -> bool:
            unique_id = f"{spider_name}_{id}"
            score = await self.r.zscore(self.processed_ids_zset, unique_id)
            return score is not None


        async def delete(self, id: str, spider_name: str) -> None:
            unique_id = f"{spider_name}_{id}"
            await self.r.delete(unique_id)
            self.logger.debug("deleted", id=unique_id, db=self.name)


        async def cleanup(self, days: int) -> None:
            threshold_timestamp = int(time.time()) - (days * 24 * 60 * 60)
            await self.r.zremrangebyscore(self.processed_ids_zset, 0, threshold_timestamp)


        async def close(self) -> None:
            if self.r:
                await self.r.aclose()

except ModuleNotFoundError:
    # proxy pattern
    class RedisDB(DBClient):
        name = "redis"
        
        _ERROR_MESSAGE = (
            f"The '{name}' component is disabled because the required dependencies are not installed. "
            "Please install it to enable this feature:\n\n"
            "  pip install 'zenx[redis]'"
        )

        def __init__(self, *args, **kwargs):
            raise ImportError(self._ERROR_MESSAGE)
        
        async def start(self) -> None: pass
        async def _connect(self) -> None: pass
        async def insert(self, id: str, spider_name: str) -> None: pass
        async def exists(self, id: str, spider_name: str) -> bool: return False
        async def delete(self, id: str, spider_name: str) -> None: pass
        async def cleanup(self, days: int) -> None: pass
        async def close(self) -> None: pass
