from __future__ import annotations
from abc import ABC, abstractmethod
from typing import ClassVar, Dict, Type
from structlog import BoundLogger

from zenx.clients.database import DBClient
from zenx.settings import Settings


class Pipeline(ABC):
    # central registry
    name: ClassVar[str]
    _registry: ClassVar[Dict[str, Type[Pipeline]]] = {}


    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "name"):
            raise TypeError(f"Pipeline subclass {cls.__name__} must have a 'name' attribute.")
        cls._registry[cls.name] = cls


    @classmethod
    def get_pipeline(cls, name: str) -> Type[Pipeline]:
        if name not in cls._registry:
            raise ValueError(f"Pipeline '{name}' is not registered. Available pipelines: {list(cls._registry.keys())}")
        return cls._registry[name]


    def __init__(self, logger: BoundLogger, db: DBClient, settings: Settings) -> None:
        self.logger = logger
        self.db = db
        self.settings = settings
        
    
    @abstractmethod
    async def start(self) -> None:
        """ connect and monitor the pipeline """
        ...


    @abstractmethod
    async def process_item(self, item: Dict, spider: str) -> Dict:
        ...


    @abstractmethod
    async def close(self) -> None:
        ...
